@@ -2,17 +2,12 @@ package JSON::Validator;
 use Mojo::Base -base;
 
 use Carp qw(confess);
-use JSON::Validator::Ref;
 use JSON::Validator::Store;
 use JSON::Validator::Util qw(E data_checksum is_type);
-use Mojo::File qw(path);
-use Mojo::URL;
-use Mojo::Util qw(monkey_patch sha1_sum);
-use Scalar::Util qw(blessed refaddr);
+use Mojo::Util qw(sha1_sum);
+use Scalar::Util qw(blessed);
 
-use constant RECURSION_LIMIT => $ENV{JSON_VALIDATOR_RECURSION_LIMIT} || 100;
-
-our $VERSION = '4.25';
+our $VERSION = '5.00';
 
 our %SCHEMAS = (
   'http://json-schema.org/draft-04/schema#'             => '+Draft4',
@@ -39,74 +34,7 @@ for my $method (qw(cache_paths ua)) {
   Mojo::Util::monkey_patch(__PACKAGE__, $method => sub { shift->store->$method(@_) });
 }
 
-sub bundle {
-  my ($self, $args) = @_;
-  my $cloner;
-
-  my $get_data  = $self->can('data') ? 'data' : 'schema';
-  my $schema    = $self->_new_schema($args->{schema} || $self->$get_data);
-  my $schema_id = $schema->id;
-  my @topics    = ([$schema->data, my $bundle = {}]);                        # ([$from, $to], ...);
-
-  if ($args->{replace}) {
-    $cloner = sub {
-      my $from      = shift;
-      my $from_type = ref $from;
-      my $tied      = $from_type eq 'HASH' && tied %$from;
-
-      $from = $tied->schema if $tied;
-      my $to = $from_type eq 'ARRAY' ? [] : $from_type eq 'HASH' ? {} : $from;
-      push @topics, [$from, $to] if $from_type;
-      return $to;
-    };
-  }
-  else {
-    $cloner = sub {
-      my $from      = shift;
-      my $from_type = ref $from;
-      my $tied      = $from_type eq 'HASH' && tied %$from;
-
-      unless ($tied) {
-        my $to = $from_type eq 'ARRAY' ? [] : $from_type eq 'HASH' ? {} : $from;
-        push @topics, [$from, $to] if $from_type;
-        return $to;
-      }
-
-      # Traverse all $ref
-      while (my $tmp = tied %{$tied->schema}) { $tied = $tmp }
-
-      return $from if !$args->{schema} and $tied->fqn =~ m!^\Q$schema_id\E\#!;
-
-      my $path = $self->_definitions_path($bundle, $tied);
-      unless ($self->{bundled_refs}{$tied->fqn}++) {
-        push @topics, [_node($schema->data, $path, 1, 0) || {}, _node($bundle, $path, 1, 1)];
-        push @topics, [$tied->schema, _node($bundle, $path, 0, 1)];
-      }
-
-      $path = join '/', '#', @$path;
-      tie my %ref, 'JSON::Validator::Ref', $tied->schema, $path;
-      return \%ref;
-    };
-  }
-
-  local $self->{bundled_refs} = {};
-
-  while (@topics) {
-    my ($from, $to) = @{shift @topics};
-    if (ref $from eq 'ARRAY') {
-      for (my $i = 0; $i < @$from; $i++) {
-        $to->[$i] = $cloner->($from->[$i]);
-      }
-    }
-    elsif (ref $from eq 'HASH') {
-      for my $key (keys %$from) {
-        $to->{$key} //= $cloner->($from->{$key});
-      }
-    }
-  }
-
-  return $bundle;
-}
+sub bundle { shift->schema->bundle(@_)->data }
 
 sub coerce {
   my $self = shift;
@@ -120,7 +48,7 @@ sub coerce {
   return $self;
 }
 
-sub get { JSON::Validator::Util::schema_extract(shift->schema->data, shift) }
+sub get { shift->schema->get(@_) }
 
 sub load_and_validate_schema {
   my ($self, $schema, $args) = @_;
@@ -150,101 +78,9 @@ sub schema {
 
 sub validate {
   my ($self, $data, $schema) = @_;
-  Mojo::Util::deprecated('validation(..., $schema) will be deprecated. Set schema() first instead') if $schema;
-
-  $schema //= $self->schema->data;
-  return E '/', 'No validation rules defined.' unless defined $schema;
-
-  local $self->{schema}      = $self->_new_schema($schema);
-  local $self->{seen}        = {};
-  local $self->{temp_schema} = [];                            # make sure random-errors.t does not fail
-  my @errors = sort { $a->path cmp $b->path } $self->_validate($_[1], '', $schema);
-  return @errors;
+  return +(defined $schema ? $self->_new_schema($schema) : $self->schema)->validate($_[1]);
 }
 
-sub _definitions_path {
-  my ($self, $bundle, $ref) = @_;
-  my $path = $self->_definitions_path_for_ref($ref);
-
-  # No need to rewrite, if it already has a nice name
-  my $node   = _node($bundle, $path, 2, 0);
-  my $prefix = join '/', @$path;
-  if ($ref->fqn =~ m!#/$prefix/([^/]+)$!) {
-    my $key = $1;
-
-    if ( $self->{bundled_refs}{$ref->fqn}
-      or !$node
-      or !$node->{$key}
-      or data_checksum($ref->schema) eq data_checksum($node->{$key}))
-    {
-      return [@$path, $key];
-    }
-  }
-
-  # Generate definitions key based on filename
-  my $fqn = Mojo::URL->new($ref->fqn);
-  my $key = $fqn->fragment;
-  if ($fqn->scheme and $fqn->scheme eq 'file') {
-    $key = join '-', map { s!^\W+!!; $_ } grep {$_} path($fqn->path)->basename, $key,
-      substr(sha1_sum($fqn->path), 0, 10);
-  }
-
-  # Fallback or nicer path name
-  $key =~ s![^\w-]!_!g;
-  return [@$path, $key];
-}
-
-sub _definitions_path_for_ref { ['definitions'] }
-
-# Try not to break JSON::Validator::OpenAPI::Mojolicious
-sub _get { shift; JSON::Validator::Util::_schema_extract(@_) }
-
-sub _find_and_resolve_refs {
-  my ($self, $base_url, $schema) = @_;
-  my %root = is_type($schema, 'HASH') ? %$schema : ();
-
-  my ($id_key, @topics, @refs, %seen) = ($self->_id_key, [$base_url, $schema]);
-  while (@topics) {
-    my ($base_url, $topic) = @{shift @topics};
-
-    if (is_type $topic, 'ARRAY') {
-      push @topics, map { [$base_url, $_] } @$topic;
-    }
-    elsif (is_type $topic, 'HASH') {
-      next if $seen{refaddr($topic)}++;
-
-      my $base_url = $base_url;    # do not change the global $base_url
-      if ($topic->{$id_key} and !ref $topic->{$id_key}) {
-        my $id = Mojo::URL->new($topic->{$id_key});
-        $id = $id->to_abs($base_url) unless $id->is_abs;
-        $self->store->add($id => $topic);
-        $base_url = $id;
-      }
-
-      my $has_ref = $topic->{'$ref'} && !ref $topic->{'$ref'} && !tied %$topic ? 1 : 0;
-      push @refs, [$base_url, $topic] if $has_ref;
-
-      for my $key (keys %$topic) {
-        next unless ref $topic->{$key};
-        next if $has_ref and $key eq '$ref';
-        push @topics, [$base_url, $topic->{$key}];
-      }
-    }
-  }
-
-  while (@refs) {
-    my ($base_url, $topic) = @{shift @refs};
-    next if is_type $topic, 'BOOL';
-    next if !$topic->{'$ref'} or ref $topic->{'$ref'};
-    my $base = Mojo::URL->new($base_url || $base_url)->fragment(undef);
-    my ($other, $ref_url, $fqn) = $self->_resolve_ref($topic->{'$ref'}, $base, \%root);
-    tie %$topic, 'JSON::Validator::Ref', $other, "$ref_url", "$fqn";
-    push @refs, [$other, $fqn];
-  }
-}
-
-sub _id_key { $_[0]->schema ? $_[0]->schema->_id_key : 'id' }
-
 sub _new_schema {
   my ($self, $source, %attrs) = @_;
   return $source if blessed $source and $source->can('specification');
@@ -260,6 +96,8 @@ sub _new_schema {
   my $store  = $self->store;
   my $schema = $loadable ? $store->get($store->load($source)) : $source;
 
+  $attrs{recursive_data_protection} //= $self->recursive_data_protection;
+
   $attrs{coerce}  ||= $self->{coerce}  if $self->{coerce};
   $attrs{formats} ||= $self->{formats} if $self->{formats};
   $attrs{specification} = $schema->{'$schema'}
@@ -283,117 +121,12 @@ sub _new_schema {
     }
   }
 
-  my $schema_class = $spec && $SCHEMAS{$spec} || 'JSON::Validator::Schema::Draft4';
+  my $schema_class = $spec && $SCHEMAS{$spec} || 'JSON::Validator::Schema';
   $schema_class =~ s!^\+(.+)$!JSON::Validator::Schema::$1!;
   confess "Could not load $schema_class: $@" unless $schema_class->can('new') or eval "require $schema_class;1";
   return $schema_class->new($source, %attrs);
 }
 
-sub _node {
-  my ($node, $path, $offset, $create) = @_;
-
-  my $n = 0;
-  while ($path->[$n]) {
-    $node->{$path->[$n]} ||= {} if $create;
-    return undef unless $node = $node->{$path->[$n]};
-    last if (++$n) + $offset >= @$path;
-  }
-
-  return $node;
-}
-
-sub _ref_to_schema {
-  my ($self, $schema) = @_;
-  return $schema if ref $schema ne 'HASH';
-
-  my @guard;
-  while (my $tied = tied %$schema) {
-    push @guard, $tied->ref;
-    confess "Seems like you have a circular reference: @guard" if @guard > RECURSION_LIMIT;
-    $schema = $tied->schema;
-    last if is_type $schema, 'BOOL';
-  }
-
-  return $schema;
-}
-
-sub _register_root_schema {
-  my ($self, $id, $schema) = @_;
-  confess "Root schema cannot have a fragment in the 'id'. ($id)" if $id =~ /\#./;
-  confess "Root schema cannot have a relative 'id'. ($id)" unless $id =~ /^\w+:/ or -e $id or $id =~ m!^/!;
-}
-
-# _resolve() method is used to convert all "id" into absolute URLs and
-# resolve all the $ref's that we find inside JSON Schema specification.
-sub _resolve {
-  my ($self, $schema, $nested) = @_;
-  return $schema if is_type $schema, 'BOOL';
-
-  my ($id_key, $id, $cached_id, $resolved) = ($self->_id_key);
-  if (ref $schema eq 'HASH') {
-    $id        = $schema->{$id_key} // '';
-    $cached_id = $self->store->exists($id);
-    $resolved  = $cached_id ? $self->store->get($cached_id) : $schema;
-  }
-  else {
-    $cached_id = $self->store->exists($id);
-    $id        = $cached_id // $self->store->load($schema);
-    $resolved  = $self->store->get($id);
-    $id        = $resolved->{$id_key} if is_type($resolved, 'HASH') and $resolved->{$id_key};
-  }
-
-  $cached_id //= '';
-  $id = Mojo::URL->new("$id");
-  $self->_register_root_schema($id => $resolved) if !$nested and "$id";
-  $self->store->add($id => $resolved)            if "$id"    and "$id" ne $cached_id;
-  $self->_find_and_resolve_refs($id => $resolved) unless $cached_id;
-
-  return $resolved;
-}
-
-sub _resolve_ref {
-  my ($self, $ref_url, $base_url, $schema) = @_;
-  $ref_url = "#$ref_url" if $ref_url =~ m!^/!;
-
-  my $fqn     = Mojo::URL->new($ref_url);
-  my $pointer = $fqn->fragment;
-  my $other;
-
-  $fqn = $fqn->to_abs($base_url) if "$base_url";
-  $other //= $self->store->get($fqn);
-  $other //= $self->store->get($fqn->clone->fragment(undef));
-  $other //= $self->_resolve($fqn->clone->fragment(undef), 1) if $fqn->is_abs && $fqn ne $base_url;
-  $other //= $schema;
-
-  if (defined $pointer and $pointer =~ m!^/!) {
-    $other = Mojo::JSON::Pointer->new($other)->get($pointer);
-    confess qq[Possibly a typo in schema? Could not find "$pointer" in "$fqn" ($ref_url)] unless defined $other;
-  }
-
-  $fqn->fragment($pointer // '');
-  return $other, $ref_url, $fqn;
-}
-
-# DEPRECATED
-sub _validate              { goto &JSON::Validator::Schema::_validate }
-sub _validate_all_of       { goto &JSON::Validator::Schema::_validate_all_of }
-sub _validate_any_of       { goto &JSON::Validator::Schema::_validate_any_of }
-sub _validate_any_of_types { goto &JSON::Validator::Schema::_validate_any_of_types }
-sub _validate_format       { goto &JSON::Validator::Schema::_validate_format }
-sub _validate_number_max   { goto &JSON::Validator::Schema::_validate_number_max }
-sub _validate_number_min   { goto &JSON::Validator::Schema::_validate_number_min }
-sub _validate_one_of       { goto &JSON::Validator::Schema::_validate_one_of }
-sub _validate_type_any     { goto &JSON::Validator::Schema::_validate_type_any }
-sub _validate_type_array   { goto &JSON::Validator::Schema::_validate_type_array }
-sub _validate_type_boolean { goto &JSON::Validator::Schema::_validate_type_boolean }
-sub _validate_type_const   { goto &JSON::Validator::Schema::_validate_type_const }
-sub _validate_type_enum    { goto &JSON::Validator::Schema::_validate_type_enum }
-sub _validate_type_integer { goto &JSON::Validator::Schema::_validate_type_integer }
-sub _validate_type_null    { goto &JSON::Validator::Schema::_validate_type_null }
-sub _validate_type_number  { goto &JSON::Validator::Schema::_validate_type_number }
-sub _validate_type_object  { goto &JSON::Validator::Schema::_validate_type_object }
-sub _validate_type_string  { goto &JSON::Validator::Schema::_validate_type_string }
-
 1;
 
 =encoding utf8
@@ -407,7 +140,8 @@ JSON::Validator - Validate data against a JSON schema
 =head2 Using a schema object
 
 L<JSON::Validator::Schema> or any of the sub classes can be used instead of
-L<JSON::Validator>.
+L<JSON::Validator>. The only reason to use L<JSON::Validator> directly is if
+you don't know the schema version up front.
 
 =head2 Basics
 
