@@ -4,13 +4,11 @@ use Mojo::Base 'JSON::Validator::Schema';
 use JSON::Validator::Schema::Draft4;
 use JSON::Validator::Schema::Draft6;
 use JSON::Validator::Schema::Draft7;
-use JSON::Validator::Util qw(E is_type json_pointer);
-use Scalar::Util qw(blessed refaddr);
-
-my $ANCHOR_RE = qr{[A-Za-z][A-Za-z0-9:._-]*};
+use JSON::Validator::URI qw(uri);
+use JSON::Validator::Util qw(E is_bool is_type);
 
+has moniker       => 'draft2019';
 has specification => 'https://json-schema.org/draft/2019-09/schema';
-has _anchors      => sub { +{} };
 
 sub _build_formats {
   my $formats = shift->JSON::Validator::Schema::Draft7::_build_formats;
@@ -19,85 +17,73 @@ sub _build_formats {
   return $formats;
 }
 
-sub _definitions_path_for_ref { ['$defs'] }
-sub _id_key                   {'$id'}
+sub _bundle_ref_path_expand  { local $_ = $_[1]; s!^\$defs/!!; return '$defs', $_; }
+sub _extract_ref_from_schema { $_[1]->{'$recursiveRef'} // $_[1]->{'$ref'} }
 
-sub _find_and_resolve_refs {
-  my ($self, $base_url, $root) = @_;
+sub _resolve_object {
+  my ($self, $state, $schema, $refs, $found) = @_;
 
-  my (@topics, @recursive_refs, @refs, %seen) = ([$base_url, $root]);
-  while (@topics) {
-    my ($base_url, $topic) = @{shift @topics};
-
-    if (is_type $topic, 'ARRAY') {
-      push @topics, map { [$base_url, $_] } @$topic;
-    }
-    elsif (is_type $topic, 'HASH') {
-      next if $seen{refaddr($topic)}++;
-
-      my $base_url = $base_url;    # do not change the global $base_url
-      if ($topic->{'$id'} and !ref $topic->{'$id'}) {
-        my $id = Mojo::URL->new($topic->{'$id'});
-        $id = $id->to_abs($base_url) unless $id->is_abs;
-        $self->store->add($id->to_string => $topic);
-        $base_url = $id;
-      }
-
-      if ($topic->{'$anchor'} && !ref $topic->{'$anchor'}) {
-        $self->_anchors->{$topic->{'$anchor'}} = $topic;
-      }
-
-      my $is_tied           = tied %$topic;
-      my $has_ref           = !$is_tied && $topic->{'$ref'}          && !ref $topic->{'$ref'}          ? 1 : 0;
-      my $has_recursive_ref = !$is_tied && $topic->{'$recursiveRef'} && !ref $topic->{'$recursiveRef'} ? 1 : 0;
-      push @refs,           [$base_url, $topic] if $has_ref;
-      push @recursive_refs, [$base_url, $topic] if $has_recursive_ref;
-
-      for my $key (keys %$topic) {
-        next unless ref $topic->{$key};
-        next if $has_ref           and $key eq '$ref';
-        next if $has_recursive_ref and $key eq '$recursiveRef';
-        push @topics, [$base_url, $topic->{$key}];
-      }
-    }
+  if ($schema->{'$id'} and !ref $schema->{'$id'}) {
+    my $id = uri $schema->{'$id'}, $state->{base_url};
+    $self->store->add($id => $schema);
+    $state = {%$state};                                 # make sure we don't mutate $state ref
+    $state->{base_url} = $id->clone->fragment(undef);
+  }
+  if ($schema->{'$anchor'} && !ref $schema->{'$anchor'}) {
+    my $id = uri(uri()->new->fragment($schema->{'$anchor'}), $state->{base_url});
+    $self->store->add($id => $schema);
+    $state = {%$state, base_url => $id->fragment(undef)->to_string};
   }
 
-  %seen = ();
-  while (@refs) {
-    my ($base_url, $topic) = @{shift @refs};
-    next if is_type $topic, 'BOOL';
-    next if !$topic->{'$ref'} or ref $topic->{'$ref'};
-    my $base = Mojo::URL->new($base_url || $base_url)->fragment(undef);
-    my ($other, $ref_url, $fqn) = $self->_resolve_ref($topic->{'$ref'}, $base, $root);
-    next if $seen{$fqn}++ and tied %$topic;
-    tie %$topic, 'JSON::Validator::Ref', $other, $topic, "$fqn";
-    push @refs, [$fqn, $other];
+  if ($found->{'$recursiveRef'} = $schema->{'$recursiveRef'} && !ref $schema->{'$recursiveRef'}) {
+    push @$refs, [$schema, $state];
+  }
+  if ($found->{'$ref'} = $schema->{'$ref'} && !ref $schema->{'$ref'}) {
+    push @$refs, [$schema, $state];
   }
 
-  %seen = ();
-  while (@recursive_refs) {
-    my ($base_url, $topic) = @{shift @recursive_refs};
-    my $base = Mojo::URL->new($base_url || $base_url)->fragment(undef);
-    my ($other, $ref_url, $fqn) = $self->_resolve_ref($topic->{'$recursiveRef'}, $base, $root);
-    next if $seen{$fqn}++ and tied %$topic;
-    tie %$topic, 'JSON::Validator::Ref', $other, $topic, "$fqn";
-  }
+  return $state;
 }
 
-sub _resolve_ref {
-  my ($self, $ref_url, $base_url, $root) = @_;
-  return $self->_anchors->{$1}, $ref_url, $ref_url if $ref_url =~ m!^#($ANCHOR_RE)$!;
-  return $self->SUPER::_resolve_ref($ref_url, $base_url, $root);
+sub _state {
+  my ($self, $curr, %override) = @_;
+  my $schema = $override{schema};
+  my (%alongside, %seen);
+
+  while (ref $schema eq 'HASH') {
+    last unless my $ref = $schema->{'$ref'} || $schema->{'$recursiveRef'};
+    last if ref $ref;
+    last if $seen{$schema}++;
+
+    %alongside = (%alongside, %$schema);
+    $schema    = $self->_refs->{$schema}{schema}
+      // Carp::confess(qq(You have to call resolve() before validate() to lookup "$ref".));
+  }
+
+  return {%$curr, %override, schema => $schema} unless ref $schema eq 'HASH';
+
+  delete $alongside{$_} for qw($anchor $id $recursiveAnchor $recursiveRef $ref);
+  return {%$curr, %override, schema => {%alongside, %$schema}};
+}
+
+sub _state_for_get {
+  my ($self, $schema, $state) = @_;
+  return $self->_refs->{$schema}
+    if ref $schema eq 'HASH'
+    and (($schema->{'$ref'} and !ref $schema->{'$ref'})
+    or ($schema->{'$recursiveRef'} and !ref $schema->{'$recursiveRef'}));
+  return {%$state, schema => $schema};
 }
 
 sub _validate_type_array_contains {
-  my ($self, $data, $path, $schema) = @_;
+  my ($self, $data, $state) = @_;
+  my ($path, $schema) = @$state{qw(path schema)};
   return unless exists $schema->{contains};
   return if defined $schema->{minContains} and $schema->{minContains} == 0;
 
   my ($n_valid, @e, @errors) = (0);
   for my $i (0 .. @$data - 1) {
-    my @tmp = $self->_validate($data->[$i], "$path/$i", $schema->{contains});
+    my @tmp = $self->_validate($data->[$i], $self->_state($state, path => [@$path, $i], schema => $schema->{contains}));
     @tmp ? push @e, \@tmp : $n_valid++;
   }
 
@@ -111,27 +97,27 @@ sub _validate_type_array_contains {
 }
 
 sub _validate_type_object_dependencies {
-  my ($self, $data, $path, $schema) = @_;
-  my $dependencies = $schema->{dependentSchemas} || {};
+  my ($self, $data, $state) = @_;
+  my $dependencies = $state->{schema}{dependentSchemas} || {};
   my @errors;
 
   for my $k (keys %$dependencies) {
     next if not exists $data->{$k};
     if (ref $dependencies->{$k} eq 'ARRAY') {
       push @errors,
-        map { E json_pointer($path, $_), [object => dependencies => $k] }
+        map { E [@{$state->{path}}, $_], [object => dependencies => $k] }
         grep { !exists $data->{$_} } @{$dependencies->{$k}};
     }
     else {
-      push @errors, $self->_validate($data, $path, $dependencies->{$k});
+      push @errors, $self->_validate($data, $self->_state($state, schema => $dependencies->{$k}));
     }
   }
 
-  $dependencies = $schema->{dependentRequired} || {};
+  $dependencies = $state->{schema}{dependentRequired} || {};
   for my $k (keys %$dependencies) {
     next if not exists $data->{$k};
     push @errors,
-      map { E json_pointer($path, $_), [object => dependencies => $k] }
+      map { E [@{$state->{path}}, $_], [object => dependencies => $k] }
       grep { !exists $data->{$_} } @{$dependencies->{$k}};
   }
 
